# How to Update the Database

## Table of contents
1. [Final Goal](#final-goal)
2. [Download](#downloading)
3. [Import](#importing-into-sql-server)
4. [Stored Procedures](#stored-procedures)
5. [Running SQL Query](#sql-query-to-export)
6. [Final Step](#final-step)

## Final Goal
The final goal of this exercise is to create a CSV file containing following entries:

```
[StartIP],[EndIP],[ISO_Country]
```
Both StartIP and EndIP are long integers and ISO_Country is a two-digit country code.



##Downloading
Go to https://dev.maxmind.com/geoip/geoip2/geolite2/ to download the file. You will have to create an account. The IP2Country file will give you three CSV files:

1. `GeoLite2-Country-Blocks-IPv4.csv` - For IPv4
2. `GeoLite2-Country-Blocks-IPv6.csv` - For IPv6
3. `GeoLite2-Country-Locations-en.csv` - Contains Country, city information

##Importing Into SQL Server
After downloading the files, import this data into SQL Server by dragging and dropping IPv4.csv file in WinSQL. The design of the table is:

###IP2Country4
Import `GeoLite2-Country-Blocks-IPv4.csv` into this table using WinSQL drag-n-drop

```sql
CREATE TABLE [dbo].[IP2Country4](
     [network] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [geoname_id] [int] NULL ,
     [registered_country_geoname_id] [int] NULL ,
     [represented_country_geoname_id] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [is_anonymous_proxy] [int] NULL ,
     [is_satellite_provider] [int] NULL 
) ON [PRIMARY]
```

###CountryNames
Import `GeoLite2-Country-Locations-en.csv` into CountryNames table, with the following design.

```sql
CREATE TABLE [dbo].[CountryNames](
     [geoname_id] [int] NULL ,
     [locale_code] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [continent_code] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [continent_name] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [country_iso_code] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [country_name] [varchar](255) COLLATE SQL_Latin1_General_CP1_CI_AS NULL ,
     [is_in_european_union] [int] NULL 
) ON [PRIMARY]

```

##Stored Procedures
You will need the following stored procedures in SQL server to help convert CIDR format into Integers.


### Converting Long to IP address

```sql 
create function dbo.ConvertLongToIp(@IpLong bigint) 
returns varchar(15) as begin
declare
     @IpHex     varchar(8)
    ,@IpDotted  varchar(15)

select
     @IpHex = substring(convert(varchar(30), master.dbo.fn_varbintohexstr(@IpLong)), 11, 8)

select
    @IpDotted = convert(varchar(3), convert(int, (convert(varbinary, substring(@IpHex, 1, 2), 2)))) + '.' +
                convert(varchar(3), convert(int, (convert(varbinary, substring(@IpHex, 3, 2), 2)))) + '.' +
                convert(varchar(3), convert(int, (convert(varbinary, substring(@IpHex, 5, 2), 2)))) + '.' +
                convert(varchar(3), convert(int, (convert(varbinary, substring(@IpHex, 7, 2), 2))))


return @IpDotted
end
```

### Converting IP to Long

```sql
/* Adapted from http://bitmugger.blogspot.com/2011/02/how-to-convert-cidr-ip-address-to-range.html */
create function [dbo].[ConvertIPToLong](@IP varchar(15))
returns bigint
as
begin
    declare @Long bigint
    set @Long = CONVERT(bigint, PARSENAME(@IP, 4)) * 256 * 256 * 256 +
        CONVERT(bigint, PARSENAME(@IP, 3)) * 256 * 256 +
        CONVERT(bigint, PARSENAME(@IP, 2)) * 256 +
        CONVERT(bigint, PARSENAME(@IP, 1))

    return (@Long)
end
```

### Get the first IP from a CIDR address
```sql
create function [dbo].[GetCidrIpAddressStart](@Cidr varchar(50))
returns bigint
as

begin
declare
     @Base      bigint      = cast(4294967295 as bigint)
    ,@IpAddress varchar(39) = left(@Cidr, patindex('%/%' , @Cidr) - 1)
    ,@Mask      int         = Cast(substring(@Cidr, patindex('%/%' , @Cidr) + 1, 2) as int)

declare @Power      bigint  = Power(2.0, 32.0 - @Mask) - 1
declare @LowRange   bigint  = dbo.ConvertIPToLong(@IpAddress) & (@Base ^ @Power)
declare @HighRange  bigint  = @LowRange + @Power
return @LowRange
end
```

### Get the last IP from a CIDR address
```sql
create function [dbo].[GetCidrIpAddressEnd](@Cidr varchar(50))
returns bigint
as

begin
declare
     @Base      bigint      = cast(4294967295 as bigint)
    ,@IpAddress varchar(39) = left(@Cidr, patindex('%/%' , @Cidr) - 1)
    ,@Mask      int         = Cast(substring(@Cidr, patindex('%/%' , @Cidr) + 1, 2) as int)

declare @Power      bigint  = Power(2.0, 32.0 - @Mask) - 1
declare @LowRange   bigint  = dbo.ConvertIPToLong(@IpAddress) & (@Base ^ @Power)
declare @HighRange  bigint  = @LowRange + @Power
return @HighRange
end
```


### Expand CIDR address
The following stored procedure is NOT used but can return a table, which expands a CIDR address.

```sql
create function [dbo].[GetCidrIpAddressRangeAlt](@Cidr varchar(50))
returns @result table 
(
    LowRange    varchar(15) not null,
    HighRange   varchar(15) not null,
    AddressQty  bigint not null
)
as

begin
declare
     @Base      bigint      = cast(4294967295 as bigint)
    ,@IpAddress varchar(39) = left(@Cidr, patindex('%/%' , @Cidr) - 1)
    ,@Mask      int         = Cast(substring(@Cidr, patindex('%/%' , @Cidr) + 1, 2) as int)

declare @Power      bigint  = Power(2.0, 32.0 - @Mask) - 1
declare @LowRange   bigint  = dbo.ConvertIPToLong(@IpAddress) & (@Base ^ @Power)
declare @HighRange  bigint  = @LowRange + @Power

insert @result
select
     LowRange   = dbo.ConvertLongToIp(@LowRange)
    ,HighRange  = dbo.ConvertLongToIp(@HighRange)
    ,AddressQty = convert(bigint, power(2.0, (32.0 - @Mask)))

return
end

```
Following example to uses the stored procedure.

```sql
select * from Customers.dbo.GetCidrIpAddressRangeAlt('1.0.16.20/24')
```
 
 
 
### SQL Query To Export
Once the data has been imported and the Stored procedures are created in the database, run the following query to export data into a CSV file

```sql
-- Final Query
-- STEP 1 - Fetch where country is known
-- Some of the records in CountryNames table have NULL for country code. Those are
-- Replaced by 'zz'

select Customers.dbo.GetCidrIpAddressStart(network) Start_IP, 
       Customers.dbo.GetCidrIpAddressEnd(network) End_IP, 
       ISNULL(country_iso_code, 'ZZ')
from IP2Country4 i JOIN CountryNames c on i.geoname_id = c.geoname_id
UNION
-- Step 2 - If country is not known, get the registered country
select Customers.dbo.GetCidrIpAddressStart(network) Start_IP, 
       Customers.dbo.GetCidrIpAddressEnd(network) End_IP, 
       ISNULL(country_iso_code, 'ZZ')
from IP2Country4 i JOIN CountryNames c on i.registered_country_geoname_id = c.geoname_id
where i.geoname_id is null
order by Start_IP
```


Export the results of this query to `ip2c.dat` file using WinSQL's Export Wizard. Ensure the field separator is a comma and enclosure character is blank.


##Final Step

Finally, zip `ip2c.dat` into `ip2c.zip` and upload it to `http://xeams.com/files/ip2c.zip`
